/*****************************************************************************
 *
 * FILE:	httpd.c
 * DESCRIPTION:	httpd server
 * DATE:	Wed, Jun 11 2003
 * UPDATED:	Mon, Sep  3 2007
 * AUTHOR:	Kouichi ABE (WALL) / 
 * E-MAIL:	kouichi@MysticWALL.COM
 * URL:		http://www.MysticWALL.COM/
 * COPYRIGHT:	(c) 2003-2007 졿Kouichi ABE (WALL), All rights reserved.
 * LICENSE:
 *
 *  Copyright (c) 2003-2007 Kouichi ABE (WALL) <kouichi@MysticWALL.COM>,
 *  All rights reserved.
 *  
 *  Redistribution and use in source and binary forms, with or without
 *  modification, are permitted provided that the following conditions
 *  are met:
 *
 *   1. Redistributions of source code must retain the above copyright
 *      notice, this list of conditions and the following disclaimer.
 *   2. Redistributions in binary form must reproduce the above copyright
 *      notice, this list of conditions and the following disclaimer in the
 *      documentation and/or other materials provided with the distribution.
 *
 *   THIS SOFTWARE IS PROVIDED BY THE AUTHOR AND CONTRIBUTORS ``AS IS'' AND
 *   ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 *   IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 *   ARE DISCLAIMED.  IN NO EVENT SHALL THE AUTHOR OR CONTRIBUTORS BE LIABLE
 *   FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
 *   DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS
 *   OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 *   HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
 *   LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY
 *   OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
 *   SUCH DAMAGE.
 *
 * $Id: httpd.c,v 1.3 2007/09/03 00:25:04 kouichi Exp $
 *
 *****************************************************************************/

#include <stdio.h>
#include <string.h>
#include <unistd.h>
#include <stdlib.h>
#include <fcntl.h>
#include <sys/types.h>
#include <sys/time.h>
#include <sys/resource.h>
#include <sys/stat.h>
#include <signal.h>
#include <errno.h>
#include "netfd.h"

/*****************************************************************************
 *
 *	Program and Copyright
 *
 *****************************************************************************/
static const char	Program[] = "httpd";
static const char	Copyright[] =
	"Copyright (c) 2003-2007 Kouichi ABE (WALL), All rights reserved.";

/*****************************************************************************
 *
 *	Macros and structures definition
 *
 *****************************************************************************/
#define	CRLF	"\015\012"
#define	LF	'\012'
#define	CR	'\015'
#define	EOL	'\0'

#define	DEF_ADDR	"192.168.1.23"
#define	DEF_PORT	"80"
#define	DEF_LOGFILE	"/dev/null"
#define	DEF_BACKLOG	(48)
#define	DEF_INDEX	"index.html"

typedef struct _ListenAddr {
  const char *		ipaddr;
  const char *		port;
  netfd_t *		nsd;
  struct _ListenAddr *	next;
} ListenAddr;

typedef struct {
  char *	addr;
  char *	port;
  char *	docdir;
  char *	cgidir;
  char *	logfile;
  uid_t		uid;
  /**/
  ListenAddr *	listen;
  /**/
  FILE *	logger;
} Httpd;

enum {
  EST400,	/* bad request */
  EST403,	/* forbidden */
  EST404,	/* not found */
  EST405,	/* method not allowed */
  EST500	/* internal server error */
};

/*****************************************************************************
 *
 *	Local functions declaration
 *
 *****************************************************************************/
static ListenAddr *	new_listen(const char *, const char *);
static void		web_child(netfd_t *);
static char *		get_uri(netfd_t *);
static void		http_response(netfd_t *, const char *);
static void		http_mime(netfd_t *, const char *);
static void		http_error(netfd_t *, int);
static void		http_date(netfd_t *);

static void		set_trap(void);
static void		trap(int);
static void		print_cpu_time(void);

/*****************************************************************************
 *
 *	Local variables declaration
 *
 *****************************************************************************/
static Httpd	httpd = {
  DEF_ADDR,	/* Address */
  DEF_PORT,	/* Port */
  NULL,		/* Document Root */
  NULL,		/* cgi-bin directory */
  DEF_LOGFILE,	/* Access/Error Log File */
  65535,	/* User */
  NULL,
  NULL
};

/*****************************************************************************
 *
 *	Functions definition
 *
 *****************************************************************************/

static ListenAddr *
new_listen(ipaddr, port)
	const char *	ipaddr;
	const char *	port;
{
  ListenAddr *	new;

  new = (ListenAddr *)malloc(sizeof(ListenAddr));
  if (new == NULL) {
    perror("malloc");
    exit(1);
  }
  new->ipaddr	= strdup(ipaddr);
  new->port	= strdup(port);
  new->nsd	= NULL;
  new->next	= httpd.listen;
  httpd.listen	= new;

  return new;
}

static void
web_child(nfd)
	netfd_t *	nfd;
{
  char *	uri;

  uri = get_uri(nfd);
  if (uri != NULL) {
    char	req[BUFSIZ];
    ssize_t	n;

    memset(req, 0, sizeof(req));
    while ((n = netfd_readline(nfd, req, sizeof(req), NETFD_NO_TIMEOUT)) > 0) {
      if (memcmp(req, CRLF, 2) == 0) {
       	break;	/* request end */
      }
      memset(req, 0, sizeof(req));
    }
    if (n < 0) {	/* error */
      http_error(nfd, EST500);
    }
    else {
      http_response(nfd, uri);
    }
    free(uri);
  }
  netfd_close(nfd);
}

static char *
get_uri(nfd)
	register netfd_t *	nfd;
{
  char		req[BUFSIZ];
  ssize_t	n;
  int		begin;
  int		end;
  size_t	len;
  char *	uri;

  memset(req, 0, sizeof(req));
  n = netfd_readline(nfd, req, sizeof(req), NETFD_NO_TIMEOUT);
  if (n <= 0) {
    http_error(nfd, EST400);
    return NULL;
  }

  /* HTTP method */
  if (strncmp(req, "GET ", 4) == 0) {
    begin = 4;
  }
  else if (strncmp(req, "HEAD ", 5) == 0) {
    begin = 5;
  }
  else {
    http_error(nfd, EST405);
    return NULL;
  }

  /* End the request of line */
  end = strstr(req, CRLF) - req;

  /* HTTP version: HTTP/x.y */
  end -= 9;
  switch (req[end + 8]) {
    case '9': break;	/* HTTP/0.9 */
    case '0': break;	/* HTTP/1.0 */
    case '1': break;	/* HTTP/1.1 */
    default:
      http_error(nfd, EST400);
      return NULL;
  }

  len = end - begin;
  uri = (char *)malloc(len + sizeof(DEF_INDEX));
  memcpy(uri, req + begin, len);
  uri[len] = EOL;
  if (uri[len - 1] == '/') {
    strlcat(uri, DEF_INDEX, len + sizeof(DEF_INDEX));
  }

  return uri;
}

static void
http_response(nsd, uri)
	register netfd_t *	nsd;
	const char *		uri;
{
  errno = 0;
  while (uri[0] == '/') {
    uri++;
  }
  if (uri != NULL) {
    netfd_t *	nfd;

    nfd = netfd_open(uri, O_RDONLY, 0400);
    if (nfd != NULL) {
      struct stat	sbuf;

      if (fstat(netfd_fileno(nfd), &sbuf) == 0) {
	char	buf[BUFSIZ];
	size_t	n;
	off_t	sbytes;

	/*
	 * send header
	 */
	n = (size_t)snprintf(buf, sizeof(buf) - 1,
			     "HTTP/1.0 200 OK\r\n"
			     "Server: %s\r\n"
			     "Content-Length: %ld\r\n",
			     Program, (long)sbuf.st_size);
	netfd_write(nsd, buf, n, NETFD_NO_TIMEOUT);
	http_date(nsd);
	n = strftime(buf, sizeof(buf),
		     "Last-Modified: %a, %d %h %Y %H:%M:%S %Z\r\n",
		     gmtime(&sbuf.st_mtime));
	netfd_write(nsd, buf, n, NETFD_NO_TIMEOUT);
	http_mime(nsd, uri);
	netfd_write(nsd, CRLF, 2, NETFD_NO_TIMEOUT);
	/*
	 * send body
	 */
	netfd_sendfile(nfd, nsd, (off_t)0, 0, NULL, &sbytes, 0, NETFD_NO_TIMEOUT);
#if	DEBUG
	fprintf(httpd.logger, "DEBUG: sendfile = %d\n", (int)sbytes);
#endif	/* DEBUG */

      }
      netfd_close(nfd);
    }
    else {
      switch (errno) {
	case EACCES: http_error(nsd, EST403); break;
	case ENOENT: http_error(nsd, EST404); break;
	default:     http_error(nsd, EST500); break;
      }
    }
  }
  else {
    http_error(nsd, EST500);
  }
}

static void
http_mime(nfd, uri)
	register netfd_t *	nfd;
	const char *		uri;
{
  const char		text_plain[] = "text/plain";
  const char		text_html[]  = "text/html";
  const char		text_css[]   = "text/css";
  const char		image_png[]  = "image/png";
  const char		image_gif[]  = "image/gif";
  const char		image_jpeg[] = "image/jpeg";
  register char *	p;

  p = strrchr(uri, '.');
  if (p != NULL) {
    struct iovec	iov[3];
    unsigned int	n = 0;

    iov[n].iov_base = "Content-Type: ";
    iov[n].iov_len  = strlen(iov[n].iov_base);
    n++;

    p++;
    if (strcmp(p, "html") == 0 || strcmp(p, "htm") == 0) {
      iov[n].iov_base = text_html;
    }
    else if (strcmp(p, "css") == 0) {
      iov[n].iov_base = text_css;
    }
    else if (strcmp(p, "png") == 0) {
      iov[n].iov_base = image_png;
    }
    else if (strcmp(p, "gif") == 0) {
      iov[n].iov_base = image_gif;
    }
    else if (strcmp(p, "jpeg") == 0 || strcmp(p, "jpg") == 0) {
      iov[n].iov_base = image_jpeg;
    }
    else {
      iov[n].iov_base = text_plain;
    }
    iov[n].iov_len  = strlen(iov[n].iov_base);
    n++;

    iov[n].iov_base = CRLF;
    iov[n].iov_len  = 2;
    n++;

    netfd_writev(nfd, iov, n, NETFD_NO_TIMEOUT);
  }
}

static void
http_error(nfd, eno)
	register netfd_t *	nfd;
	int			eno;
{
  struct {
    const unsigned int	code;
    const char *	phrase;
    const char *	body;
  } emsg[] = {
    { 400, "Bad Request", "Your request has bad syntax or is inherently impossible to satisfy." },
    { 403, "Forbidden", "You do not have permission to get the request URL." },
    { 404, "Not Found", "Not found - file does not exist or is read protected." },
    { 405, "Method Not Allowed", "This server answer only to GET request." },
    { 500, "Internal Server Error", "There was an unusual problem serving the requested URL." }
  };
  char	buf[256];
  int	n;

  n = snprintf(buf, sizeof(buf) - 1,
	       "HTTP/1.0 %u %s\r\n"
	       "Server: %s\r\n"
	       "Content-Type: text/html\r\n",
	       emsg[eno].code, emsg[eno].phrase, Program);
  netfd_write(nfd, buf, (size_t)n, NETFD_NO_TIMEOUT);
  http_date(nfd);
  netfd_write(nfd, CRLF, 2, NETFD_NO_TIMEOUT);
  n = snprintf(buf, sizeof(buf) - 1,
	       "<html><head><title>%s</title></head><body>%s</body></html>",
	       emsg[eno].phrase, emsg[eno].body);
  netfd_write(nfd, buf, (size_t)n, NETFD_NO_TIMEOUT);
}

static void
http_date(nfd)
	register netfd_t *	nfd;
{
  time_t	now = time(NULL);
  char		date[80];

  strftime(date, sizeof(date), "Date: %a, %d %h %Y %H:%M:%S %Z\r\n",
	   gmtime(&now));
  netfd_write(nfd, date, strlen(date), NETFD_NO_TIMEOUT);
}


static void
set_trap(void)
{
  struct sigaction	act;

  /* quit */
  sigemptyset(&act.sa_mask);
  act.sa_flags	 = 0;
  act.sa_handler = (void *)trap;
  sigaction(SIGINT,  &act, NULL);
  sigaction(SIGQUIT, &act, NULL);
  sigaction(SIGTERM, &act, NULL);
  /**/
  sigaction(SIGBUS,  &act, NULL);
  sigaction(SIGSEGV, &act, NULL);
}

static void
print_cpu_time(void)
{
  double	user;
  double	sys;
  struct rusage	myusage;
  struct rusage	childusage;

  if (getrusage(RUSAGE_SELF, &myusage) < 0) {
    perror("getrusage");
    exit(1);
  }
  if (getrusage(RUSAGE_CHILDREN, &childusage) < 0) {
    perror("getrusage");
    exit(1);
  }

  user = (double)myusage.ru_utime.tv_sec + myusage.ru_utime.tv_usec / 1000000.0;
  user += (double)childusage.ru_utime.tv_sec
				+ childusage.ru_utime.tv_usec / 1000000.0;
  sys = (double)myusage.ru_stime.tv_sec + myusage.ru_stime.tv_usec / 1000000.0;
  sys += (double)childusage.ru_stime.tv_sec
				+ childusage.ru_stime.tv_usec / 1000000.0;
  fprintf(httpd.logger, "user time = %g, sys time = %g\n", user, sys);
}

static void
trap(signo)
	int	signo;
{
  register ListenAddr *	lap;

  /* print statistics */
  print_cpu_time();

  for (lap = httpd.listen; lap != NULL; lap = lap->next) {
    netfd_close(lap->nsd);
  }
  fprintf(httpd.logger, "killed by sig#%d\n", signo);
  fclose(httpd.logger);
  exit(0);
}

static void
usage(void)
{
  fprintf(stderr,
	  "%s\n"
	  "usage: %s [-a addr][-p port][-d docdir][-l logfile][-u uid]\n",
	  Copyright, Program);
  exit(64);
}

int
main(argc, argv)
	int	argc;
	char *	argv[];
{
  int			ch;
  register ListenAddr *	lap;

  httpd.uid = getuid();
  while ((ch = getopt(argc, argv, "a:p:d:c:l:u:vh?")) != -1) {
    switch (ch) {
      case 'a': httpd.addr	= strdup(optarg); break;
      case 'p': httpd.port	= strdup(optarg); break;
      case 'd': httpd.docdir	= strdup(optarg); break;
      case 'c': httpd.cgidir	= strdup(optarg); break;
      case 'l': httpd.logfile	= strdup(optarg); break;
      case 'u': httpd.uid	= (uid_t)atoi(optarg); break;
      case '?':
      default:
	usage();
    }
  }
  argc -= optind;
  argv += optind;

  if (httpd.docdir == NULL) {
    httpd.docdir = getenv("HOME");
  }
  if (chdir(httpd.docdir) == -1) {
    perror("chdir");
    exit(1);
  }

  if (netfd_init(0) == -1) {
    perror("netfd_init");
    exit(1);
  }

  httpd.logger = fopen(httpd.logfile, "a");
  if (httpd.logger == NULL) {
    perror("fopen");
    exit(1);
  }
  setvbuf(httpd.logger, (char *)NULL, _IOLBF, 0);

  new_listen("127.0.0.1", httpd.port);
  new_listen(httpd.addr,  httpd.port);

  set_trap();

  for (lap = httpd.listen; lap != NULL; lap = lap->next) {
    lap->nsd = netfd_tcp_server(lap->ipaddr, lap->port, DEF_BACKLOG);
    if (lap->nsd == NULL) {
      perror("netfd_tcp_server");
      exit(1);
    }
    netfd_read_event_add(lap->nsd);
  }

  netfd_dispatch(web_child, false, -1);

  return 0;	/* not reached */
}
